@@ -100,6 +100,10 @@ decode_ie(<<?RAND, Len, Rand:Len/binary, Tail/binary>>, Map) ->
   ?CHECK_LEN(rand, Len, 16, 16),
   decode_ie(Tail, Map#{rand => Rand});
 
+decode_ie(<<?PCO, Len, PCO:Len/binary, Tail/binary>>, Map) ->
+  ?CHECK_LEN(pco, Len, 1, 251),
+  decode_ie(Tail, Map#{pco => PCO});
+
 decode_ie(<<?AUTS, Len, AUTS:Len/binary, Tail/binary>>, Map) ->
   ?CHECK_LEN(auts, Len, 14, 14),
   decode_ie(Tail, Map#{auts => AUTS});
@@ -408,6 +412,11 @@ encode_ie(#{pdp_charging := Value} = GSUPMessage, Head) ->
   ?CHECK_SIZE(pdp_charging, Len, Value),
   encode_ie(maps:without([pdp_charging], GSUPMessage), <<Head/binary, ?PDP_CHARGING, Len, Value:Len/unit:8>>);
 
+encode_ie(#{pco := Value} = GSUPMessage, Head) ->
+  Len = size(Value),
+  ?CHECK_LEN(pco, Len, 1, 251),
+  encode_ie(maps:without([pco], GSUPMessage), <<Head/binary, ?PCO, Len, Value/binary>>);
+
 encode_ie(#{auts := Value} = GSUPMessage, Head) ->
   Len = 14,
   ?CHECK_LEN(auts, size(Value), Len, Len),
