@@ -18,15 +18,15 @@
 % GNU General Public License for more details.
 %
 % You should have received a copy of the GNU Affero General Public License
-% along with this program.  If not, see <http://www.gnu.org/licenses/>.
+% along with this program.  If not, see <https://www.gnu.org/licenses/>.
 %
 % Additional Permission under GNU AGPL version 3 section 7:
 %
 % If you modify this Program, or any covered work, by linking or
 % combining it with runtime libraries of Erlang/OTP as released by
-% Ericsson on http://www.erlang.org (or a modified version of these
+% Ericsson on https://www.erlang.org (or a modified version of these
 % libraries), containing parts covered by the terms of the Erlang Public
-% License (http://www.erlang.org/EPLICENSE), the licensors of this
+% License (https://www.erlang.org/EPLICENSE), the licensors of this
 % Program grant you additional permission to convey the resulting work
 % without the need to license the runtime libraries of Erlang/OTP under
 % the GNU Affero General Public License. Corresponding Source for a
@@ -48,7 +48,7 @@
 %% gen_server Function Exports
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2]).
 -export([code_change/3]).
--export([create_session_req/3, delete_session_req/1]).
+-export([create_session_req/4, delete_session_req/1]).
 
 %% Application Definitions
 -define(SERVER, ?MODULE).
@@ -88,6 +88,8 @@
     imsi                   :: binary(),
     pid                    :: pid(),
     apn                    :: binary(),
+    raddr_str              :: string(),
+    raddr                  :: inet:ip_address(),
     ue_ip                  :: inet:ip_address(),
     local_control_tei = 0  :: non_neg_integer(),
     remote_control_tei = 0 :: non_neg_integer(),
@@ -142,15 +144,21 @@ init(State) ->
             lager:error("GTPv2C UDP socket open error: ~w~n", [Reason])
     end.
 
-create_session_req(Imsi, Apn, APCO) ->
-    gen_server:call(?SERVER, {gtpc_create_session_req, {Imsi, Apn, APCO}}).
+create_session_req(Imsi, Apn, APCO, PGWAddrCandidateList) ->
+    gen_server:call(?SERVER, {gtpc_create_session_req, {Imsi, Apn, APCO, PGWAddrCandidateList}}).
 
 delete_session_req(Imsi) ->
     gen_server:call(?SERVER, {gtpc_delete_session_req, {Imsi}}).
 
-handle_call({gtpc_create_session_req, {Imsi, Apn, APCO}}, {Pid, _Tag} = _From, State0) ->
+handle_call({gtpc_create_session_req, {Imsi, Apn, APCO, PGWAddrCandidateList}}, {Pid, _Tag} = _From, State0) ->
+    RemoteAddrStr = pick_gtpc_remote_address(PGWAddrCandidateList, State0),
+    lager:debug("Selected PGW Remote Address ~p~n", [RemoteAddrStr]),
+    {ok, RemoteAddrInet} = inet_parse:address(RemoteAddrStr),
     {Sess0, State1} = find_or_new_gtp_session(Imsi,
-                        #gtp_session{pid = Pid, apn = list_to_binary(Apn)},
+                        #gtp_session{pid = Pid,
+                                     apn = list_to_binary(Apn),
+                                     raddr_str = RemoteAddrInet,
+                                     raddr = RemoteAddrInet},
                         State0),
     Req = gen_create_session_request(Sess0, APCO, State1),
     tx_gtp(Req, State1),
@@ -180,10 +188,10 @@ handle_cast(Req, State) ->
 %% @callback gen_server
 handle_info({udp, _Socket, IP, InPortNo, RxMsg}, State) ->
     lager:debug("S2b: Rx from IP ~p port ~p: ~p~n", [IP, InPortNo, RxMsg]),
-    spawn_wait_ret(fun() ->
-                        rx_udp(IP, InPortNo, RxMsg, State)
-                   end,
-                   {noreply, State});
+    misc:spawn_wait_ret(fun() ->
+                            rx_udp(IP, InPortNo, RxMsg, State)
+                        end,
+                        {noreply, State});
 handle_info(Info, State) ->
     lager:info("S2b handle_info: ~p ~n", [Info]),
     {noreply, State}.
@@ -352,6 +360,16 @@ find_unused_local_teid(State) ->
     find_unused_local_teid(State, State#gtp_state.next_local_data_tei,
                            dec_tei(State#gtp_state.next_local_data_tei)).
 
+pick_gtpc_remote_address(PGWAddrCandidateList, State) ->
+    case PGWAddrCandidateList of
+    [] ->
+        %% Pick default address from configuration:
+        State#gtp_state .raddr_str;
+    [Head|_Tail] ->
+        % TODO: pick a compatible address with .laddr from PGWAddrCandidateList is exists.
+        Head
+    end.
+
 %% connect/2
 connect(Name, {Socket, RemoteAddr, RemotePort}) ->
     lager:info("~s connecting to IP ~s port ~p~n", [Name, RemoteAddr, RemotePort]),
@@ -360,22 +378,6 @@ connect(Name, {Socket, RemoteAddr, RemotePort}) ->
 connect(Address) ->
     connect(?SVC_NAME, Address).
 
-%% Calls Fun on a spawned monitored process and returns Fun ret.
-%% If spawned process crashes, return DefaultRet.
-spawn_wait_ret(Fun, DefaultRet) ->
-    MyPID=self(),
-    {Pid, MRef} = spawn_monitor(fun() ->
-                                    Ret = Fun(),
-                                    MyPID ! {self(), Ret}
-                                end),
-    receive
-    {'DOWN', MRef, process, _, _Reason} ->
-        DefaultRet;
-    {Pid, Ret} ->
-        erlang:demonitor(MRef, [flush]),
-        Ret
-    end.
-
 rx_udp(IP, InPortNo, RxMsg, State) ->
     Req = gtp_packet:decode(RxMsg),
     lager:debug("S2b: Rx from IP ~p port ~p: ~p~n", [IP, InPortNo, Req]),
@@ -389,46 +391,13 @@ rx_gtp(Resp = #gtp{version = v2, type = create_session_response}, State0) ->
             {noreply, State0};
         Sess0 ->
             % Do GTP specific msg parsing here, pass only relevant fields:
-            #{{v2_fully_qualified_tunnel_endpoint_identifier,1} :=
-                #v2_fully_qualified_tunnel_endpoint_identifier{
-                    interface_type = 32, %% "S2b PGW GTP-C"
-                    key = RemoteTEIC, ipv4 = _IPc4, ipv6 = _IPc6},
-              {v2_pdn_address_allocation,0} := Paa,
-              {v2_bearer_context,0} := #v2_bearer_context{instance = 0, group = BearerIE}} = Resp#gtp.ie,
-            % Parse BearerContext:
-            #{{v2_eps_bearer_id,0} := #v2_eps_bearer_id{instance = 0, eps_bearer_id = Ebi},
-              {v2_fully_qualified_tunnel_endpoint_identifier,4} :=
-                #v2_fully_qualified_tunnel_endpoint_identifier{
-                    interface_type = 33, %% "S2b-U PGW GTP-U"
-                    key = RemoteTEID, ipv4 = IPu4, ipv6 = IPu6}
-             } = BearerIE,
-            Bearer = gtp_session_find_bearer_by_ebi(Sess0, Ebi),
-            Sess1 = gtp_session_update_bearer(Sess0, Bearer, Bearer#gtp_bearer{remote_data_tei = RemoteTEID}),
-            Sess2 = Sess1#gtp_session{remote_control_tei = RemoteTEIC},
-            lager:info("s2b: Updated Session after create_session_response: ~p~n", [Sess2]),
-            State1 = update_gtp_session(Sess0, Sess2, State0),
-            case maps:find({v2_additional_protocol_configuration_options,0}, Resp#gtp.ie) of
-            {ok, APCO_dec} ->
-                lager:debug("s2b: APCO_dec: ~p~n", [APCO_dec]),
-                APCO = gtp_packet:encode_protocol_config_opts(APCO_dec#v2_additional_protocol_configuration_options.config);
-            error ->
-                lager:notice("s2b: APCO not found in CreateSessionResp!~n", []),
-                APCO = undefined
-            end,
-            ResInfo0 = #{
-                apn => binary_to_list(Sess0#gtp_session.apn),
-                eua => conv:gtp2_paa_to_epdg_eua(Paa),
-                local_teid => Bearer#gtp_bearer.local_data_tei,
-                remote_teid => RemoteTEID,
-                remote_ipv4 => IPu4,
-                remote_ipv6 => IPu6
-            },
-            case APCO of
-            undefined -> ResInfo = ResInfo0;
-            _ -> ResInfo = maps:put(apco, APCO, ResInfo0)
-            end,
-            epdg_ue_fsm:received_gtpc_create_session_response(Sess0#gtp_session.pid, {ok, ResInfo}),
-            {noreply, State1}
+            % First lookup Cause:
+            #{{v2_cause,0} := #v2_cause{instance = 0, v2_cause = GtpCauseAtom}} = Resp#gtp.ie,
+            GtpCause = gtp_utils:enum_v2_cause(GtpCauseAtom),
+            case gtp_utils:v2_cause_successful(GtpCause) of
+            true -> rx_gtp_create_session_response_successful(Resp, Sess0, State0);
+            false -> rx_gtp_create_session_response_failure(GtpCause, Sess0, State0)
+            end
         end;
 
 rx_gtp(Resp = #gtp{version = v2, type = delete_session_response}, State0) ->
@@ -487,6 +456,54 @@ rx_gtp(Req, State) ->
     lager:error("S2b: UNIMPLEMENTED Rx: ~p~n", [Req]),
     {noreply, State}.
 
+
+rx_gtp_create_session_response_successful(Resp, Sess0, State0) ->
+    #{{v2_fully_qualified_tunnel_endpoint_identifier,1} :=
+        #v2_fully_qualified_tunnel_endpoint_identifier{
+            interface_type = 32, %% "S2b PGW GTP-C"
+            key = RemoteTEIC, ipv4 = _IPc4, ipv6 = _IPc6},
+      {v2_pdn_address_allocation,0} := Paa,
+      {v2_bearer_context,0} := #v2_bearer_context{instance = 0, group = BearerIE}} = Resp#gtp.ie,
+    % Parse BearerContext:
+    #{{v2_eps_bearer_id,0} := #v2_eps_bearer_id{instance = 0, eps_bearer_id = Ebi},
+      {v2_fully_qualified_tunnel_endpoint_identifier,4} :=
+        #v2_fully_qualified_tunnel_endpoint_identifier{
+            interface_type = 33, %% "S2b-U PGW GTP-U"
+            key = RemoteTEID, ipv4 = IPu4, ipv6 = IPu6}
+     } = BearerIE,
+    Bearer = gtp_session_find_bearer_by_ebi(Sess0, Ebi),
+    Sess1 = gtp_session_update_bearer(Sess0, Bearer, Bearer#gtp_bearer{remote_data_tei = RemoteTEID}),
+    Sess2 = Sess1#gtp_session{remote_control_tei = RemoteTEIC},
+    lager:info("s2b: Updated Session after create_session_response: ~p~n", [Sess2]),
+    State1 = update_gtp_session(Sess0, Sess2, State0),
+    case maps:find({v2_additional_protocol_configuration_options,0}, Resp#gtp.ie) of
+    {ok, APCO_dec} ->
+        lager:debug("s2b: APCO_dec: ~p~n", [APCO_dec]),
+        APCO = gtp_packet:encode_protocol_config_opts(APCO_dec#v2_additional_protocol_configuration_options.config);
+    error ->
+        lager:notice("s2b: APCO not found in CreateSessionResp!~n", []),
+        APCO = undefined
+    end,
+    ResInfo0 = #{
+        apn => binary_to_list(Sess0#gtp_session.apn),
+        eua => conv:gtp2_paa_to_epdg_eua(Paa),
+        local_teid => Bearer#gtp_bearer.local_data_tei,
+        remote_teid => RemoteTEID,
+        remote_ipv4 => IPu4,
+        remote_ipv6 => IPu6
+    },
+    case APCO of
+    undefined -> ResInfo = ResInfo0;
+    _ -> ResInfo = maps:put(apco, APCO, ResInfo0)
+    end,
+    epdg_ue_fsm:received_gtpc_create_session_response(Sess0#gtp_session.pid, {ok, ResInfo}),
+    {noreply, State1}.
+
+rx_gtp_create_session_response_failure(GtpCause, Sess, State0) ->
+    epdg_ue_fsm:received_gtpc_create_session_response(Sess#gtp_session.pid, {error, GtpCause}),
+    State1 = delete_gtp_session(Sess, State0),
+    {noreply, State1}.
+
 tx_gtp(Req, State) ->
     lager:info("s2b: Tx ~p~n", [Req]),
     Msg = gtp_packet:encode(Req),
