@@ -14,15 +14,15 @@
 % GNU General Public License for more details.
 %
 % You should have received a copy of the GNU Affero General Public License
-% along with this program.  If not, see <http://www.gnu.org/licenses/>.
+% along with this program.  If not, see <https://www.gnu.org/licenses/>.
 %
 % Additional Permission under GNU AGPL version 3 section 7:
 %
 % If you modify this Program, or any covered work, by linking or
 % combining it with runtime libraries of Erlang/OTP as released by
-% Ericsson on http://www.erlang.org (or a modified version of these
+% Ericsson on https://www.erlang.org (or a modified version of these
 % libraries), containing parts covered by the terms of the Erlang Public
-% License (http://www.erlang.org/EPLICENSE), the licensors of this
+% License (https://www.erlang.org/EPLICENSE), the licensors of this
 % Program grant you additional permission to convey the resulting work
 % without the need to license the runtime libraries of Erlang/OTP under
 % the GNU Affero General Public License. Corresponding Source for a
@@ -42,22 +42,37 @@
 -export([start/1, stop/1]).
 -export([init/1,callback_mode/0,terminate/3]).
 -export([get_server_name_by_imsi/1, get_pid_by_imsi/1]).
--export([auth_request/2, lu_request/1, tunnel_request/2, purge_ms_request/1]).
--export([received_swm_auth_response/2, received_swm_auth_compl_response/2, received_swm_session_termination_answer/2]).
+-export([auth_request/2, lu_request/1, tunnel_request/2, purge_ms_request/1,
+         cancel_location_result/1]).
+-export([received_swm_reauth_request/1, received_swm_dea_auth_response/2,
+         received_swm_dea_auth_compl_response/2, received_swm_auth_answer/2,
+         received_swm_session_termination_answer/2, received_swm_abort_session_request/1]).
 -export([received_gtpc_create_session_response/2, received_gtpc_delete_session_response/2, received_gtpc_delete_bearer_request/1]).
--export([state_new/3, state_wait_auth_resp/3, state_authenticating/3, state_authenticated/3,
-         state_wait_create_session_resp/3, state_wait_delete_session_resp/3,
-         state_wait_swm_session_termination_answer/3, state_active/3]).
+-export([state_new/3,
+         state_wait_auth_resp/3,
+         state_authenticating/3,
+         state_authenticated/3,
+         state_wait_create_session_resp/3,
+         state_active/3,
+         state_wait_swm_session_termination_answer/3,
+         state_wait_delete_session_resp/3,
+         state_dereg_pgw_initiated_wait_cancel_location_res/3,
+         state_dereg_net_initiated_wait_cancel_location_res/3,
+         state_dereg_net_initiated_wait_s2b_delete_session_resp/3]).
 
 -define(TIMEOUT_VAL_WAIT_GTP_ANSWER, 10000).
+-define(TIMEOUT_VAL_WAIT_GSUP_ANSWER, 10000).
+-define(TIMEOUT_VAL_WAIT_SWm_ANSWER, 10000).
 
 -record(ue_fsm_data, {
         imsi,
         apn                     = "internet"    :: string(),
         pgw_rem_addr_list       = []            :: list(),
-        tun_pdp_ctx                             :: epdg_tun_pdp_ctx,
+        tun_pdp_ctx                             :: #epdg_tun_pdp_ctx{},
         tear_down_gsup_needed   = false         :: boolean(), %% need to send GSUP PurgeMSResp after STR+STA?
-        tear_down_gsup_cause    = 0             :: integer()
+        tear_down_gsup_cause    = 0             :: integer(),
+        tear_down_s2b_needed    = false         :: boolean(), %% need to send S2b DeleteSessionReq
+        tear_down_tx_swm_asa_needed = false         :: boolean() %% need to send SWm ASA
         }).
 
 get_server_name_by_imsi(Imsi) ->
@@ -117,19 +132,46 @@ purge_ms_request(Pid) ->
                 {error, Err}
         end.
 
-received_swm_auth_response(Pid, Result) ->
-        lager:info("ue_fsm received_swm_auth_response ~p~n", [Result]),
+cancel_location_result(Pid) ->
+        lager:info("ue_fsm cancel_location_result~n", []),
         try
-        gen_statem:call(Pid, {received_swm_auth_response, Result})
+                gen_statem:call(Pid, cancel_location_result)
         catch
         exit:Err ->
                 {error, Err}
         end.
 
-received_swm_auth_compl_response(Pid, Result) ->
-        lager:info("ue_fsm received_swm_auth_compl_response ~p~n", [Result]),
+received_swm_reauth_request(Pid) ->
+        lager:info("ue_fsm received_swm_reauth_request~n", []),
         try
-        gen_statem:call(Pid, {received_swm_auth_compl_response, Result})
+        gen_statem:call(Pid, received_swm_reauth_request)
+        catch
+        exit:Err ->
+                {error, Err}
+        end.
+
+received_swm_auth_answer(Pid, Result) ->
+lager:info("ue_fsm received_swm_auth_answer~n", []),
+try
+gen_statem:call(Pid, {received_swm_auth_answer, Result})
+catch
+exit:Err ->
+        {error, Err}
+end.
+
+received_swm_dea_auth_response(Pid, Result) ->
+        lager:info("ue_fsm received_swm_dea_auth_response ~p~n", [Result]),
+        try
+        gen_statem:call(Pid, {received_swm_dea_auth_response, Result})
+        catch
+        exit:Err ->
+                {error, Err}
+        end.
+
+received_swm_dea_auth_compl_response(Pid, Result) ->
+        lager:info("ue_fsm received_swm_dea_auth_compl_response ~p~n", [Result]),
+        try
+        gen_statem:call(Pid, {received_swm_dea_auth_compl_response, Result})
         catch
         exit:Err ->
                 {error, Err}
@@ -144,6 +186,15 @@ received_swm_session_termination_answer(Pid, Result) ->
                 {error, Err}
         end.
 
+received_swm_abort_session_request(Pid) ->
+        lager:info("ue_fsm received_swm_abort_session_request~n", []),
+        try
+        gen_statem:call(Pid, received_swm_asr)
+        catch
+        exit:Err ->
+                {error, Err}
+        end.
+
 received_gtpc_create_session_response(Pid, Result) ->
         lager:info("ue_fsm received_gtpc_create_session_response ~p~n", [Result]),
         try
@@ -177,10 +228,8 @@ received_gtpc_delete_bearer_request(Pid) ->
 %% ------------------------------------------------------------------
 
 ev_handle({call, From}, {auth_request, PdpTypeNr, Apn, EAP}, Data) ->
-        case epdg_diameter_swm:auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn, EAP) of
-        ok -> {next_state, state_wait_auth_resp, Data, [{reply,From,ok}]};
-        {error, Err} -> {stop_and_reply, Err, [{reply,From,{error,Err}}], Data}
-	end.
+        epdg_diameter_swm:tx_der_auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn, EAP),
+        {next_state, state_wait_auth_resp, Data, [{reply,From,ok}]}.
 
 %% ------------------------------------------------------------------
 %% gen_statem Function Definitions
@@ -202,6 +251,10 @@ terminate(Reason, State, Data) ->
         end,
         ok.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_new:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_new(enter, _OldState, Data) ->
         {keep_state, Data};
 
@@ -213,11 +266,15 @@ state_new({call, From}, purge_ms_request, Data) ->
         lager:info("ue_fsm state_new event=purge_ms_request, ~p~n", [Data]),
         {stop_and_reply, purge_ms_request, [{reply,From,ok}], Data}.
 
-state_wait_auth_resp(enter, _OldState, Data) ->
-        {keep_state, Data};
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_wait_auth_resp:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
-state_wait_auth_resp({call, From}, {received_swm_auth_response, Result}, Data) ->
-        lager:info("ue_fsm state_wait_auth_resp event=received_swm_auth_response Result=~p, ~p~n", [Result, Data]),
+state_wait_auth_resp(enter, _OldState, Data) ->
+        {keep_state, Data, {state_timeout,?TIMEOUT_VAL_WAIT_SWm_ANSWER,swm_der_timeout}};
+
+state_wait_auth_resp({call, From}, {received_swm_dea_auth_response, Result}, Data) ->
+        lager:info("ue_fsm state_wait_auth_resp event=received_swm_dea_auth_response Result=~p, ~p~n", [Result, Data]),
         case Result of
                 {ok, _AuthTuples} ->
                         gsup_server:auth_response(Data#ue_fsm_data.imsi, Result),
@@ -228,7 +285,17 @@ state_wait_auth_resp({call, From}, {received_swm_auth_response, Result}, Data) -
                         {next_state, state_new, Data, [{reply,From,ok}]};
                 _ ->
                         {next_state, state_new, Data, [{reply,From,{error,unknown}}]}
-        end.
+        end;
+
+state_wait_auth_resp(state_timeout, swm_der_timeout, Data) ->
+        lager:error("ue_fsm state_wait_auth_resp: Timeout ~p, ~p~n", [swm_der_timeout, Data]),
+        GsupCause = ?GSUP_CAUSE_NET_FAIL,
+        gsup_server:auth_response(Data#ue_fsm_data.imsi, {error, GsupCause}),
+        {next_state, state_new, Data}.
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_authenticating:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 state_authenticating(enter, _OldState, Data) ->
         {keep_state, Data};
@@ -240,14 +307,12 @@ state_authenticating({call, _From} = EvType, {auth_request, PdpTypeNr, Apn, EAP}
 state_authenticating({call, From}, lu_request, Data) ->
         lager:info("ue_fsm state_authenticating event=lu_request, ~p~n", [Data]),
         % Rx "GSUP CEAI LU Req" is our way of saying Rx "Swm Diameter-EAP REQ (DER) with EAP AVP containing successuful auth":
-        case epdg_diameter_swm:auth_compl_request(Data#ue_fsm_data.imsi, Data#ue_fsm_data.apn) of
-        ok -> {keep_state, Data, [{reply,From,ok}]};
-        {error, Err} -> {stop_and_reply, Err, [{reply,From,{error,Err}}], Data}
-        end;
+        epdg_diameter_swm:tx_der_auth_compl_request(Data#ue_fsm_data.imsi, Data#ue_fsm_data.apn),
+        {keep_state, Data, [{reply,From,ok}]};
 
 % Rx Swm Diameter-EAP Answer (DEA) containing APN-Configuration, triggered by
 % earlier Tx DER EAP AVP containing successuful auth", when we received GSUP LU Req:
-state_authenticating({call, From}, {received_swm_auth_compl_response, Result}, Data) ->
+state_authenticating({call, From}, {received_swm_dea_auth_compl_response, Result}, Data) ->
         lager:info("ue_fsm state_authenticating event=lu_request, ~p, ~p~n", [Result, Data]),
         % Rx "GSUP CEAI LU Req" is our way of saying Rx "Swm Diameter-EAP REQ (DER) with EAP AVP containing successuful auth":
         case Result of
@@ -267,6 +332,10 @@ state_authenticating({call, From}, {received_swm_auth_compl_response, Result}, D
                 {next_state, state_new, Data, [{reply,From,ok}]}
         end.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_authenticated:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_authenticated(enter, _OldState, Data) ->
         {keep_state, Data};
 
@@ -282,6 +351,29 @@ state_authenticated({call, From}, {tunnel_request, PCO}, Data) ->
                                          Data#ue_fsm_data.pgw_rem_addr_list),
         {next_state, state_wait_create_session_resp, Data, [{reply,From,ok}]};
 
+state_authenticated({call, From}, received_swm_reauth_request, Data) ->
+        lager:info("ue_fsm state_authenticated event=received_swm_reauth_request, ~p~n", [Data]),
+        epdg_diameter_swm:tx_reauth_answer(Data#ue_fsm_data.imsi, #epdg_dia_rc{result_code = 2001}),
+        % 3GPP TS 29.273 7.1.2.5.1:
+        % Upon receiving the re-authorization request, the ePDG shall immediately invoke the authorization procedure
+        % specified in 7.1.2.2 for the session indicated in the request. This procedure is based on the Diameter
+        % commands AA-Request (AAR) and AA-Answer (AAA) specified in IETF RFC 4005 [4]. Information
+        % element contents for these messages are shown in tables 7.1.2.2.1/1 and 7.1.2.2.1/2.
+        epdg_diameter_swm:tx_auth_req(Data#ue_fsm_data.imsi),
+        {keep_state, Data, [{reply,From,ok}]};
+
+
+state_authenticated({call, From}, {received_swm_auth_answer, Result}, Data) ->
+        lager:info("ue_fsm state_authenticated event=received_swm_auth_answer(~p), ~p~n", [Result, Data]),
+        case Result of
+        ok ->
+                {keep_state, Data, [{reply,From,ok}]};
+        _ ->
+                Data1 = Data#ue_fsm_data{tear_down_gsup_needed = false,
+                                         tear_down_s2b_needed = false},
+                {next_state, state_dereg_net_initiated_wait_cancel_location_res, Data1, [{reply,From,ok}]}
+        end;
+
 state_authenticated({call, From}, purge_ms_request, Data) ->
         lager:info("ue_fsm state_authenticated event=purge_ms_request, ~p~n", [Data]),
         Data1 = Data#ue_fsm_data{tear_down_gsup_needed = true},
@@ -289,9 +381,8 @@ state_authenticated({call, From}, purge_ms_request, Data) ->
 
 state_authenticated({call, From}, received_gtpc_delete_bearer_request, Data) ->
         lager:info("ue_fsm state_authenticated event=received_gtpc_delete_bearer_request, ~p~n", [Data]),
-        gsup_server:cancel_location_request(Data#ue_fsm_data.imsi),
         Data1 = Data#ue_fsm_data{tear_down_gsup_needed = false},
-        {next_state, state_wait_swm_session_termination_answer, Data1, [{reply,From,ok}]};
+        {next_state, state_dereg_pgw_initiated_wait_cancel_location_res, Data1, [{reply,From,ok}]};
 
 state_authenticated({call, From}, Event, Data) ->
         lager:error("ue_fsm state_authenticated: Unexpected call event ~p, ~p~n", [Event, Data]),
@@ -301,6 +392,10 @@ state_authenticated(cast, Event, Data) ->
         lager:error("ue_fsm state_authenticated: Unexpected cast event ~p, ~p~n", [Event, Data]),
         {keep_state, Data}.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_wait_create_session_resp:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_wait_create_session_resp(enter, _OldState, Data) ->
         {keep_state, Data, [{state_timeout,?TIMEOUT_VAL_WAIT_GTP_ANSWER,create_session_timeout}]};
 
@@ -327,7 +422,7 @@ state_wait_create_session_resp({call, From}, {received_gtpc_create_session_respo
         end;
 
 state_wait_create_session_resp({call, From}, Event, Data) ->
-        lager:error("ue_fsm state_wait_delete_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
+        lager:error("ue_fsm state_wait_create_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
         {keep_state, Data, [{reply,From,{error,unexpected_event}}]};
 
 state_wait_create_session_resp(state_timeout, create_session_timeout, Data) ->
@@ -335,6 +430,10 @@ state_wait_create_session_resp(state_timeout, create_session_timeout, Data) ->
         gsup_server:tunnel_response(Data#ue_fsm_data.imsi, {error, ?GSUP_CAUSE_CONGESTION}),
         {next_state, state_authenticated, Data}.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_active:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_active(enter, _OldState, Data) ->
         {keep_state, Data};
 
@@ -344,6 +443,31 @@ state_active({call, _From}, {auth_request, PdpTypeNr, Apn, EAP}, Data) ->
         Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined},
         {next_state, state_new, Data1, [postpone]};
 
+state_active({call, From}, received_swm_reauth_request, Data) ->
+        lager:info("ue_fsm state_active event=received_swm_reauth_request, ~p~n", [Data]),
+        epdg_diameter_swm:tx_reauth_answer(Data#ue_fsm_data.imsi, #epdg_dia_rc{result_code = 2001}),
+        % 3GPP TS 29.273 7.1.2.5.1:
+        % Upon receiving the re-authorization request, the ePDG shall immediately invoke the authorization procedure
+        % specified in 7.1.2.2 for the session indicated in the request. This procedure is based on the Diameter
+        % commands AA-Request (AAR) and AA-Answer (AAA) specified in IETF RFC 4005 [4]. Information
+        % element contents for these messages are shown in tables 7.1.2.2.1/1 and 7.1.2.2.1/2.
+        epdg_diameter_swm:tx_auth_req(Data#ue_fsm_data.imsi),
+        {keep_state, Data, [{reply,From,ok}]};
+
+state_active({call, From}, {received_swm_auth_answer, Result}, Data) ->
+        lager:info("ue_fsm state_active event=received_swm_auth_answer(~p), ~p~n", [Result, Data]),
+        case Result of
+        ok ->
+                {keep_state, Data, [{reply,From,ok}]};
+        _ ->
+                gtp_u_tun:delete_pdp_context(Data#ue_fsm_data.tun_pdp_ctx),
+                Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined,
+                                         tear_down_gsup_needed = false,
+                                         tear_down_s2b_needed = true,
+                                         tear_down_tx_swm_asa_needed = false},
+                {next_state, state_dereg_net_initiated_wait_cancel_location_res, Data1, [{reply,From,ok}]}
+        end;
+
 state_active({call, From}, purge_ms_request, Data) ->
         lager:info("ue_fsm state_active event=purge_ms_request, ~p~n", [Data]),
         gtp_u_tun:delete_pdp_context(Data#ue_fsm_data.tun_pdp_ctx),
@@ -356,9 +480,21 @@ state_active({call, From}, purge_ms_request, Data) ->
 state_active({call, From}, received_gtpc_delete_bearer_request, Data) ->
         lager:info("ue_fsm state_active event=received_gtpc_delete_bearer_request, ~p~n", [Data]),
         gtp_u_tun:delete_pdp_context(Data#ue_fsm_data.tun_pdp_ctx),
-        gsup_server:cancel_location_request(Data#ue_fsm_data.imsi),
-        Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined, tear_down_gsup_needed = false},
-        {next_state, state_wait_swm_session_termination_answer, Data1, [{reply,From,ok}]};
+        Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined,
+                                 tear_down_gsup_needed = false,
+                                 tear_down_s2b_needed = false,
+                                 tear_down_tx_swm_asa_needed = false},
+        {next_state, state_dereg_pgw_initiated_wait_cancel_location_res, Data1, [{reply,From,ok}]};
+
+%%% network (HSS/AAA) initiated de-registation requested:
+state_active({call, From}, received_swm_asr, Data) ->
+        lager:info("ue_fsm state_active event=received_swm_asr, ~p~n", [Data]),
+        gtp_u_tun:delete_pdp_context(Data#ue_fsm_data.tun_pdp_ctx),
+        Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined,
+                                 tear_down_gsup_needed = false,
+                                 tear_down_s2b_needed = true,
+                                 tear_down_tx_swm_asa_needed = true},
+        {next_state, state_dereg_net_initiated_wait_cancel_location_res, Data1, [{reply,From,ok}]};
 
 state_active({call, From}, Event, Data) ->
         lager:error("ue_fsm state_active: Unexpected call event ~p, ~p~n", [Event, Data]),
@@ -368,6 +504,10 @@ state_active(cast, Event, Data) ->
         lager:error("ue_fsm state_active: Unexpected cast event ~p, ~p~n", [Event, Data]),
         {keep_state, Data}.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_wait_delete_session_resp:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_wait_delete_session_resp(enter, _OldState, Data) ->
         {keep_state, Data};
 
@@ -388,19 +528,42 @@ state_wait_delete_session_resp({call, From}, Event, Data) ->
         lager:error("ue_fsm state_wait_delete_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
         {keep_state, Data, [{reply,From,{error,unexpected_event}}]}.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_dereg_pgw_initiated_wait_cancel_location_res:
+%% Network (PGW) initiated de-registration: We trigger GSUP Cancel Location Req.
+%% Wait for GSUP Cancel Location Result.
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+state_dereg_pgw_initiated_wait_cancel_location_res(enter, _OldState, Data) ->
+        case gsup_server:cancel_location_request(Data#ue_fsm_data.imsi) of
+        ok ->
+                {keep_state, Data, {state_timeout,?TIMEOUT_VAL_WAIT_GSUP_ANSWER,gsup_cancel_location_timeout}};
+        {error, _Err} ->
+                {next_state, state_wait_swm_session_termination_answer, Data}
+        end;
+
+state_dereg_pgw_initiated_wait_cancel_location_res({call, From}, cancel_location_result, Data) ->
+        lager:info("ue_fsm state_dereg_pgw_initiated_wait_cancel_location_res event=cancel_location_result, ~p~n", [Data]),
+        {next_state, state_wait_swm_session_termination_answer, Data, [{reply,From,ok}]};
+
+state_dereg_pgw_initiated_wait_cancel_location_res({call, From}, Event, Data) ->
+        lager:error("ue_fsm state_dereg_pgw_initiated_wait_cancel_location_res: Unexpected call event ~p, ~p~n", [Event, Data]),
+        {keep_state, Data, [{reply,From,ok}]};
+
+
+state_dereg_pgw_initiated_wait_cancel_location_res(state_timeout, gsup_cancel_location_timeout, Data) ->
+        lager:error("ue_fsm state_dereg_pgw_initiated_wait_cancel_location_res: Timeout ~p, ~p~n", [s2b_delete_session_timeout, Data]),
+        {next_state, state_wait_swm_session_termination_answer, Data}.
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_wait_swm_session_termination_answer:
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 state_wait_swm_session_termination_answer(enter, _OldState, Data) ->
         % Send STR towards AAA-Server
         % % 3GPP TS 29.273 7.1.2.3
         lager:info("ue_fsm state_wait_swm_session_termination_answer event=enter, ~p~n", [Data]),
-        case epdg_diameter_swm:session_termination_request(Data#ue_fsm_data.imsi) of
-        ok -> {keep_state, Data};
-        {error, _Err} ->
-                case Data#ue_fsm_data.tear_down_gsup_needed of
-                true -> gsup_server:purge_ms_response(Data#ue_fsm_data.imsi, {error, ?GSUP_CAUSE_NET_FAIL});
-                false -> ok
-                end,
-                {keep_state, Data}
-        end;
+        epdg_diameter_swm:tx_session_termination_request(Data#ue_fsm_data.imsi),
+        {keep_state, Data, {state_timeout,?TIMEOUT_VAL_WAIT_SWm_ANSWER,swm_str_timeout}};
 
 state_wait_swm_session_termination_answer({call, From}, {received_swm_sta, DiaRC}, Data) ->
         lager:info("ue_fsm state_wait_swm_session_termination_answer event=received_swm_sta, ~p~n", [Data]),
@@ -416,5 +579,84 @@ state_wait_swm_session_termination_answer({call, From}, {received_swm_sta, DiaRC
         {stop_and_reply, normal, [{reply,From,ok}], Data};
 
 state_wait_swm_session_termination_answer({call, From}, Event, Data) ->
-        lager:error("ue_fsm state_wait_delete_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
-        {keep_state, Data, [{reply,From,{error,unexpected_event}}]}.
+        lager:error("ue_fsm state_wait_swm_session_termination_answer: Unexpected call event ~p, ~p~n", [Event, Data]),
+        {keep_state, Data, [{reply,From,{error,unexpected_event}}]};
+
+state_wait_swm_session_termination_answer(state_timeout, swm_str_timeout, Data) ->
+        case Data#ue_fsm_data.tear_down_gsup_needed of
+        true -> gsup_server:purge_ms_response(Data#ue_fsm_data.imsi, {error, ?GSUP_CAUSE_NET_FAIL});
+        false -> ok
+        end,
+        {stop, normal}.
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_dereg_net_initiated_wait_cancel_location_res:
+%% Network (AAA/HSS) initiated de-registration: We trigger GSUP Cancel Location Req.
+%% Wait for GSUP Cancel Location Result, then continue to tear down S2b session.
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+state_dereg_net_initiated_wait_cancel_location_res(enter, _OldState, Data) ->
+        case gsup_server:cancel_location_request(Data#ue_fsm_data.imsi, ?GSUP_CANCELLATION_TYPE_WITHDRAW) of
+        ok ->
+                {keep_state, Data, {state_timeout,?TIMEOUT_VAL_WAIT_GSUP_ANSWER,gsup_cancel_location_timeout}};
+        {error, _Err} ->
+                {next_state, state_dereg_net_initiated_wait_s2b_delete_session_resp, Data}
+        end;
+
+state_dereg_net_initiated_wait_cancel_location_res({call, From}, cancel_location_result, Data) ->
+        lager:info("ue_fsm state_dereg_net_initiated_wait_cancel_location_res event=cancel_location_result, ~p~n", [Data]),
+        {next_state, state_dereg_net_initiated_wait_s2b_delete_session_resp, Data, [{reply,From,ok}]};
+
+state_dereg_net_initiated_wait_cancel_location_res({call, From}, Event, Data) ->
+        lager:error("ue_fsm state_dereg_net_initiated_wait_cancel_location_res: Unexpected call event ~p, ~p~n", [Event, Data]),
+        {keep_state, Data, [{reply,From,ok}]};
+
+
+state_dereg_net_initiated_wait_cancel_location_res(state_timeout, gsup_cancel_location_timeout, Data) ->
+        lager:error("ue_fsm state_dereg_net_initiated_wait_cancel_location_res: Timeout ~p, ~p~n", [s2b_delete_session_timeout, Data]),
+        {next_state, state_dereg_net_initiated_wait_s2b_delete_session_resp, Data}.
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% state_dereg_net_initiated_wait_s2b_delete_session_resp:
+%% Network (AAA/HSS) initiated de-registration: We have informed UE (GSUP), and
+%% have triggered GTPCv1 Delete Session Req against PGW.
+%% Wait for GTPCv1 Delete Session Response, ssend SWm ASA to AAAA and terminate FSM.
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+tx_swm_asa_if_needed(Data) ->
+        case Data#ue_fsm_data.tear_down_tx_swm_asa_needed of
+        true ->
+                epdg_diameter_swm:tx_abort_session_answer(Data#ue_fsm_data.imsi);
+        false -> lager:debug("Skip sending SWm ASA", [])
+        end.
+
+state_dereg_net_initiated_wait_s2b_delete_session_resp(enter, _OldState, Data) ->
+        case Data#ue_fsm_data.tear_down_s2b_needed of
+        true ->
+                case epdg_gtpc_s2b:delete_session_req(Data#ue_fsm_data.imsi) of
+                ok ->
+                        {keep_state, Data, {state_timeout,?TIMEOUT_VAL_WAIT_GTP_ANSWER,s2b_delete_session_timeout}};
+                {error, Err} ->
+                        tx_swm_asa_if_needed(Data),
+                        {stop, {error,Err}}
+                end;
+        false ->
+                tx_swm_asa_if_needed(Data),
+                {stop, normal}
+        end;
+
+state_dereg_net_initiated_wait_s2b_delete_session_resp({call, From}, {received_gtpc_delete_session_response, _Resp = #gtp{version = v2, type = delete_session_response, ie = IEs}}, Data) ->
+        lager:info("ue_fsm state_dereg_net_initiated_wait_s2b_delete_session_resp event=received_gtpc_delete_session_response, ~p~n", [Data]),
+        #{{v2_cause,0} := CauseIE} = IEs,
+        GtpCause = gtp_utils:enum_v2_cause(CauseIE#v2_cause.v2_cause),
+        lager:debug("Cause: GTP_atom=~p -> GTP_int=~p~n", [CauseIE#v2_cause.v2_cause, GtpCause]),
+        tx_swm_asa_if_needed(Data),
+        {stop_and_reply, normal, [{reply,From,ok}], Data};
+
+state_dereg_net_initiated_wait_s2b_delete_session_resp({call, From}, Event, Data) ->
+        lager:error("ue_fsm state_dereg_net_initiated_wait_s2b_delete_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
+        {keep_state, Data, [{reply,From,ok}]};
+
+
+state_dereg_net_initiated_wait_s2b_delete_session_resp(state_timeout, s2b_delete_session_timeout, Data) ->
+        lager:error("ue_fsm state_dereg_net_initiated_wait_s2b_delete_session_resp: Timeout ~p, ~p~n", [s2b_delete_session_timeout, Data]),
+        tx_swm_asa_if_needed(Data),
+        {stop, normal}.
